# Built-in/Generic Imports
from typing import Union, Optional, Any

# Exceptions
from fexception import FAttributeError, FTypeError

__author__ = "IncognitoCoding"
__copyright__ = "Copyright 2022, type_check"
__credits__ = ["IncognitoCoding"]
__license__ = "MIT"
__version__ = "0.0.9"
__maintainer__ = "IncognitoCoding"
__status__ = "Beta"


def type_check(
    value: Any,
    required_type: Union[type, tuple[type, ...]],
    enforce: bool = True,
    tb_remove_name: Optional[str] = None,
    msg_override: Optional[str] = None,
) -> Union[FTypeError, bool]:
    """
    A simple type validation check. This function is designed to be widely used to check any values.

    Raises a cleanly formatted reason if the type validation is unsuccessful.

    A None value will return different exceptions based on msg_override.\\
    \t\\- No msg_override = FAttributeError\\
    \t\\- msg_override = FTypeError

    Calling Examples:
    \t\\- type_check('sample', (int, str))\\
    \t\\- type_check('sample', str)

    NOTE:
    \t\\- Pylance type detection does not work through type_checks.\\
    \t\t\\- It would be best if you used isinstance for Pylance type checking.

    Args:
        value (any):
        \t\\- Any value needing its type validated.\\
        required_type (Union[type, tuple[type, ...]]):
        \t\\- The required type the value should match.\\
        \t\\- Can be a single type or typle of types.
        enforce (bool, optional):
        \t\\- A failed match will return a raised FTypeError exception.\\
        \t\\- Defaults to True.
        tb_remove_name (str, optional):\\
        \t\\- Caller function name or any other function in the\\
        \t   traceback chain.\\
        \t\\- Removes all traceback before and at this function.\\
        \t\\- Defaults to None.
        msg_override (str, optional):
        \t\\- Main top-level message override.\\
        \t\\- The expected and returned results will be the same.\\
        \t\\- Ideal for type checks for other importing files such as YAML.\\
        \t\\- Defaults to None.

    Raises:
        FAttributeError (fexception):
        \t\\- The value \'{value}\' sent is not an accepted input.
        FTypeError (fexception):
        \t\\- A user defined msg_override message.
        FAttributeError (fexception):
        \t\\- No type or list of types has been entered for type validation.
        FAttributeError (fexception):
        \t\\- The enforce argument value is not the correct type.
        FTypeError (fexception):
        \t\\- The value \'{value}\' is not in {required_type} format.
    
    Returns:
        Union[FTypeError, bool]:
        \t\\- FTypeError:\\
        \t\t\\- Returns raised formatted exception of the type check match fails\\
        \t\t   when 'enforce' is set to True.\\
        \t\\- bool:
        \t\t\\- Returns True if the type matches.\\
        \t\t\\- False will return when 'enforce' is set to False.
    """
    # Verifies a value is sent.
    if value is None or value == "":
        # Sets message override if one is provided.
        if msg_override:
            exc_args: dict = {
                "main_message": msg_override,
                "expected_result": "Any value other than None or an empty string",
                "returned_result": type(value),
            }
            if not tb_remove_name:
                tb_remove_name = "type_check"
            raise FTypeError(message_args=exc_args, tb_limit=None, tb_remove_name=tb_remove_name)
        else:
            exc_args: dict = {
                "main_message": f"The value '{value}' sent is not an accepted input.",
                "expected_result": "Any value other than None or an empty string",
                "returned_result": type(value),
            }
            if not tb_remove_name:
                tb_remove_name = "type_check"
            raise FAttributeError(message_args=exc_args, tb_limit=None, tb_remove_name=tb_remove_name)

    # Verifies a type or tuple is sent.
    if not (isinstance(required_type, type) or isinstance(required_type, tuple)):
        exc_args: dict = {
            "main_message": "No type or tuple of types has been entered for type validation.",
            "expected_result": "type or typle of types",
            "returned_result": type(required_type),
        }
        if not tb_remove_name:
            tb_remove_name = "type_check"
        raise FAttributeError(message_args=exc_args, tb_limit=None, tb_remove_name=tb_remove_name)

    # Verifies a bool is sent.
    if not isinstance(enforce, bool):
        exc_args: dict = {
            "main_message": "The enforce argument value is not the correct type.",
            "expected_result": "<class 'bool'>",
            "returned_result": type(enforce),
        }
        if not tb_remove_name:
            tb_remove_name = "type_check"
        raise FAttributeError(message_args=exc_args, tb_limit=None, tb_remove_name=tb_remove_name)

    matching_type_flag: bool = False
    if isinstance(value, required_type):
        matching_type_flag = True
    else:
        # Gives the option to return standard bool or an exception.
        if enforce:
            matching_type_flag = False
        else:
            return False

    # Checks for no match.
    if matching_type_flag is False:
        adjusted_required_type: Union[str, type]
        # Converts the list to a string with an '|' value between each type.
        if isinstance(required_type, tuple):
            adjusted_required_type = " | ".join(map(str, required_type))
        else:
            adjusted_required_type = required_type

        # Sets message override if one is provided.
        if msg_override:
            main_message: str = msg_override
        else:
            main_message: str = (
                f"The object value '{value}' is not an instance of the required class(es) or subclass(es)."
            )

        exc_args: dict = {
            "main_message": main_message,
            "expected_result": adjusted_required_type,
            "returned_result": type(value),
        }
        if not tb_remove_name:
            tb_remove_name = "type_check"
        raise FTypeError(message_args=exc_args, tb_limit=None, tb_remove_name=tb_remove_name)
    else:
        return True



__author__ = "IncognitoCoding"
__copyright__ = "Copyright 2022, exceptions"
__credits__ = ["IncognitoCoding"]
__license__ = "MIT"
__version__ = "0.0.1"
__maintainer__ = "IncognitoCoding"
__status__ = "Beta"


class InputFailure(Exception):
    """Exception raised for an input exception message."""

    __module__ = "builtins"
    pass


class InvalidKeyError(Exception):
    """Exception raised for an invalid dictionary key."""

    __module__ = "builtins"
    pass



from setuptools import setup

if __name__ == "__main__":
    try:
        setup()
    except:  # noqa
        print(
            "\n\nAn error occurred while building the fexception project.\n\n"
            "Please check that your setuptools version is up to date.\n"
            "Verify Command: pip install -U setuptools setuptools_scm wheel\n\n"
        )
        raise



# Built-in/Generic Imports
from typing import Union, Any

# Local Functions
from ..type.type_checks import type_check

# Local Exceptions
from ..exceptions.exceptions import InvalidKeyError

# Exceptions
from fexception import FCustomException, FAttributeError

__author__ = "IncognitoCoding"
__copyright__ = "Copyright 2022, KeyCheck"
__credits__ = ["IncognitoCoding"]
__license__ = "MIT"
__version__ = "0.0.9"
__maintainer__ = "IncognitoCoding"
__status__ = "Beta"


class KeyCheck:
    """
    An advanced dictionary key checker that offers two different check options.

    Raises a cleanly formatted reason if the key validation is unsuccessful.

    No return output.

    Options:\\
    \t\\- contains_keys():\\
    \t\t\\- Checks if some required keys exist in the dictionary.\\
    \t\\- all_keys():\\
    \t\t\\- Checks if all required keys exist in the dictionary.

    Args:
        values (dict):\\
        \t\\- A dictionary that needs the keys validated.
        \t\t\\- A template can be used with the reverse option enabled.
        enforce (bool, optional):
        \t\\- A failed match will return a raised FTypeError exception.\\
        \t\\- Defaults to True.
    
    Raises:
        FTypeError (fexception):
        \t\\- The object value '{values}' is not an instance of the required class(es) or subclass(es).
        FTypeError (fexception):
        \t\\- The object value '{enforce}' is not an instance of the required class(es) or subclass(es).
    """

    def __init__(
        self, values: dict[Union[int, float, bool, str, tuple[Union[int, float, bool, str]]], Any], enforce: bool = True
    ) -> None:
        type_check(value=values, required_type=dict, tb_remove_name="__init__")
        type_check(value=enforce, required_type=bool, tb_remove_name="__init__")

        self.__values: dict = values
        self.__required_keys: Union[Any, tuple[Any, ...]] = ""
        self.__enforce: bool = enforce

    def contains_keys(
        self,
        required_keys: Union[
            int,
            float,
            bool,
            str,
            tuple[Union[int, float, bool, str]],
            tuple[Union[int, float, bool, str, tuple[Union[int, float, bool, str]]], ...],
        ],
        reverse_output: bool = False,
    ) -> Union[InvalidKeyError, bool]:
        """
        Checks if some required keys exist in the dictionary.

        Args:
            required_keys (Union[int, float, bool, str, tuple[Union[int, float, bool, str]],\\
            \t\t\t\t   tuple[Union[int, float, bool, str, tuple[Union[int, float,\\
            \t\t\t\t\t\t\t\t\t\t\t\t\t bool, str]]], ...]):\\
            \t\\- The required key(s) that should match.\\
            \t\\- Can be a single str or tuple of keys.
            reverse (bool, optional):\\
            \t\\- Reverses the key check exception output, so the\\
            \t   expected result and returned results are flipped.\\
            \t\\- Defaults to False.

        Raises:
            FTypeError (fexception):
            \t\\- The required dictionary key(s) (required_keys) must be immutable data types.
            FTypeError (fexception):
            \t\\- The object value '{reverse_output}' is not an instance of the required class(es) or subclass(es).

        Nested Raises:
            FAttributeError (fexception):\\
            \t\\- No key(s) were sent.
            FAttributeError (fexception):\\
            \t\\- The input keys have inconsistent value and requirement keys.\\
            FAttributeError (fexception):\\
            \t\\- The required key tuple contains duplicate keys. All keys must be unique.\\
            InvalidKeyError (fexception):\\
            \t\\- The dictionary key (\'{non_matching_key}\')\\
            \t  does not exist in the expected required key(s).
        
        Returns:
            Union[InvalidKeyError, bool]:
            \t\\- InvalidKeyError:\\
            \t\t\\- Returns raised formatted exception of the type check match fails\\
            \t\t   when 'enforce' is set to True.\\
            \t\\- bool:
            \t\t\\- Returns True if the type matches.\\
            \t\t\\- False will return when 'enforce' is set to False.
        """
        type_check(
            value=required_keys,
            required_type=(int, float, bool, str, tuple),
            tb_remove_name="contains_keys",
            msg_override="The required dictionary key(s) (required_keys) must be immutable data types.",
        )
        type_check(value=reverse_output, required_type=bool, tb_remove_name="contains_keys")

        self.__required_keys = required_keys
        self.__all_key_check: bool = False
        self.__reverse_output: bool = reverse_output

        return self.__key_validation()

    def all_keys(
        self,
        required_keys: Union[
            int,
            float,
            bool,
            str,
            tuple[Union[int, float, bool, str]],
            tuple[Union[int, float, bool, str, tuple[Union[int, float, bool, str]]], ...],
        ],
        reverse_output: bool = False,
    ) -> Union[InvalidKeyError, bool]:
        """
        Checks if all required keys exist in the dictionary

        Args:
            required_keys (Union[int, float, bool, str, tuple[Union[int, float, bool, str]],\\
            \t\t\t\t   tuple[Union[int, float, bool, str, tuple[Union[int, float,\\
            \t\t\t\t\t\t\t\t\t\t\t\t\t bool, str]]], ...]):\\
            \t\\- The required key(s) that should match.\\
            \t\\- Can be a single str or tuple of keys.
            reverse (bool, optional):\\
            \t\\- Reverses the key check exception output, so the\\
            \t   expected result and returned results are flipped.\\
            \t\\- Defaults to False.

        Raises:
            FTypeError (fexception):
            \t\\- The required dictionary key(s) (required_keys) must be immutable data types.
            FTypeError (fexception):
            \t\\- The object value '{reverse_output}' is not an instance of the required class(es) or subclass(es).

        Nested Raises:
            FAttributeError (fexception):\\
            \t\\- No key(s) were sent.
            FAttributeError (fexception):\\
            \t\\- The input keys have inconsistent value and requirement keys.\\
            FAttributeError (fexception):\\
            \t\\- The required key tuple contains duplicate keys. All keys must be unique.\\
            InvalidKeyError (fexception):\\
            \t\\- The dictionary key (\'{non_matching_key}\')\\
            \t  does not exist in the expected required key(s).
        
        Returns:
            Union[InvalidKeyError, bool]:
            \t\\- InvalidKeyError:\\
            \t\t\\- Returns raised formatted exception of the type check match fails\\
            \t\t   when 'enforce' is set to True.\\
            \t\\- bool:
            \t\t\\- Returns True if the type matches.\\
            \t\t\\- False will return when 'enforce' is set to False.
        """
        type_check(
            value=required_keys,
            required_type=(int, float, bool, str, tuple),
            tb_remove_name="all_keys",
            msg_override="The required dictionary key(s) (required_keys) must be immutable data types.",
        )
        if reverse_output:
            type_check(value=reverse_output, required_type=bool, tb_remove_name="all_keys")

        self.__required_keys = required_keys
        self.__all_key_check: bool = True
        self.__reverse_output: bool = reverse_output

        return self.__key_validation()

    def __key_validation(self) -> Union[InvalidKeyError, bool]:
        """
        Performs the key validation.

        Raises:
            FAttributeError (fexception):\\
            \t\\- No key(s) were sent.
            FAttributeError (fexception):\\
            \t\\- The input keys have inconsistent value and requirement keys.\\
            FAttributeError (fexception):\\
            \t\\- The required key tuple contains duplicate keys. All keys must be unique.\\
            InvalidKeyError (fexception):\\
            \t\\- The dictionary key (\'{non_matching_key}\')\\
            \t  does not exist in the expected required key(s).
        
        Returns:
            Union[InvalidKeyError, bool]:
            \t\\- InvalidKeyError:\\
            \t\t\\- Returns raised formatted exception of the file is not found.\\
            \t\\- bool:
            \t\t\\- Returns True if the key(s) are valid.\\
            \t\t\\- No False bool will return. This is replaced by the raised exception.
        """
        # Loops through to find any keys that do not match.
        dict_keys = tuple(self.__values.keys())

        # Reverses key results for flipped reverse checks.
        required_key_result: Union[Any, tuple]
        expected_key_result: Union[Any, tuple]
        if self.__reverse_output:
            expected_key_result = dict_keys
            required_key_result = self.__required_keys
        else:
            expected_key_result = self.__required_keys
            required_key_result = dict_keys

        # Checks for the required keys are sent.
        if not self.__required_keys:
            # Formats the output based on the check option.
            if self.__all_key_check:
                expected_result: str = f"  - Expected Key(s) = {expected_key_result}"
            else:
                expected_result: str = f"  - Expected Match Option Key(s) = {dict_keys}"

            exc_args = {
                "main_message": "No key(s) were sent.",
                "expected_result": expected_result,
                "returned_result": None,
            }
            if self.__all_key_check is True:
                raise FAttributeError(message_args=exc_args, tb_limit=None, tb_remove_name="all_keys")
            else:
                raise FAttributeError(message_args=exc_args, tb_limit=None, tb_remove_name="contains_keys")

        # Checks for 1:1 input when using the all_keys option.
        mismatched_input: bool = False
        if self.__all_key_check:
            mismatched_input: bool
            if isinstance(self.__required_keys, tuple):
                if len(dict_keys) != len(self.__required_keys):
                    mismatched_input = True
                else:
                    mismatched_input = False
            else:
                if len(self.__values) > 1:
                    mismatched_input = True

            if mismatched_input is True:
                exc_args: dict = {
                    "main_message": "The input keys have inconsistent value and requirement keys.",
                    "expected_result": f"Required Key(s) = {expected_key_result}",
                    "returned_result": f"Failed Key(s) = {required_key_result}",
                }
                raise FAttributeError(message_args=exc_args, tb_limit=None, tb_remove_name="all_keys")

        # Checks for duplicate values.
        if isinstance(self.__required_keys, tuple):
            if len(self.__required_keys) != len(set(self.__required_keys)):
                exc_args: dict = {
                    "main_message": "The required key tuple contains duplicate keys. All keys must be unique.",
                    "returned_result": f"Required Key(s) = {self.__required_keys}",
                }
                if self.__all_key_check is True:
                    raise FAttributeError(message_args=exc_args, tb_limit=None, tb_remove_name="all_keys")
                else:
                    raise FAttributeError(message_args=exc_args, tb_limit=None, tb_remove_name="contains_keys")

        # Sets the keys in reverse order so the no-match is the last entry checked
        # but the first no-match in the tuple of keys.
        sorted_dict_keys = sorted(dict_keys, reverse=True)

        non_matching_key: Union[Any, None] = None
        if isinstance(self.__required_keys, tuple):
            for required_key in self.__required_keys:
                # Checks if the validation requires all the required keys
                # to match all sorted_dict_keys or the required keys to match
                # some of the sorted_dict_keys.
                if self.__all_key_check:
                    for dict_key in sorted_dict_keys:
                        # Checks for exact match.
                        if required_key == dict_key:
                            non_matching_key = None
                            break
                        else:
                            non_matching_key = required_key
                else:
                    if required_key in sorted_dict_keys:
                        non_matching_key = None
                    else:
                        non_matching_key = required_key
                # If a match is not found on the first required
                # key check the loop will exit and return the no-matched key.
                if non_matching_key:
                    break
        else:
            # Variable name swap for easier loop reading.
            required_key: str = self.__required_keys
            for dict_key in sorted_dict_keys:
                if required_key == dict_key:
                    # Checks for exact match.
                    non_matching_key = None
                    break
                else:
                    non_matching_key = required_key

        # Checks if a no matching key exists, to output the error
        if non_matching_key and not self.__enforce:
            return False
        elif non_matching_key and self.__enforce:
            # Formats the output based on the check option.
            if self.__all_key_check:
                main_message: str = (
                    f"The dictionary key ('{non_matching_key}') " "does not exist in the expected required key(s).\n"
                )
                expected_result: str = f"Expected Key(s) = {expected_key_result}"
                returned_result: str = f"Failed Key(s) = {required_key_result}"
            else:
                main_message: str = (
                    f"The dictionary key ('{non_matching_key}') " "does not match any expected match option key(s).\n"
                )
                if isinstance(expected_key_result, tuple):
                    expected_result: str = f"Match Option Key(s) = {expected_key_result}"
                else:
                    expected_result: str = f"Match Option Key(s) = '{expected_key_result}'"
                returned_result: str = f"Failed Key(s) = {required_key_result}"

            exc_args: dict = {
                "main_message": main_message,
                "custom_type": InvalidKeyError,
                "expected_result": expected_result,
                "returned_result": returned_result,
            }
            if self.__all_key_check is True:
                raise InvalidKeyError(FCustomException(message_args=exc_args, tb_limit=None, tb_remove_name="all_keys"))
            else:
                raise InvalidKeyError(
                    FCustomException(message_args=exc_args, tb_limit=None, tb_remove_name="contains_keys")
                )
        else:
            return True



# Built-in/Generic Imports
import pathlib
from typing import Optional, Union

# Local Functions
from ..type.type_checks import type_check

# Exceptions
from fexception import FFileNotFoundError


__author__ = "IncognitoCoding"
__copyright__ = "Copyright 2022, file_checks"
__credits__ = ["IncognitoCoding"]
__license__ = "MIT"
__version__ = "0.0.9"
__maintainer__ = "IncognitoCoding"
__status__ = "Beta"


def file_check(
    file_path: str, file_description: Optional[str] = None, enforce: bool = True
) -> Union[FFileNotFoundError, bool]:
    """
    Checks if the file exists.

    Raises a cleanly formatted reason if the key validation is unsuccessful.

    Args:
        file_path (str):
        \t\\- The file path being checked.
        file_description (str, optional):
        \t\\- Name of the file being checked.\\
        \t\\- Defaults to None.
        enforce (bool, optional):
        \t\\- A failed match will return a raised FTypeError exception.\\
        \t\\- Defaults to True.

    Raises:
        FTypeError (fexception):
        \t\\- The object value '{file_path}' is not an instance of the required class(es) or subclass(es).
        FTypeError (fexception):
        \t\\- The object value '{file_description}' is not an instance of the required class(es) or subclass(es).
        FTypeError (fexception):
        \t\\- The object value '{enforce}' is not an instance of the required class(es) or subclass(es).
        FFileNotFoundError (fexception):
        \t\\- The file ({file_description}) does not exist in the validating file path ({file_path}).
        FFileNotFoundError (fexception):
        \t\\- The file does not exist in the validating file path ({file_path}).

    Returns:
        Union[FFileNotFoundError, bool]:
        \t\\- FFileNotFoundError:\\
        \t\t\\- Returns raised formatted exception of the type check match fails\\
        \t\t   when 'enforce' is set to True.\\
        \t\\- bool:
        \t\t\\- Returns True if the type matches.\\
        \t\t\\- False will return when 'enforce' is set to False.
    """
    type_check(value=file_path, required_type=str, tb_remove_name="file_checks")
    if file_description:
        type_check(value=file_description, required_type=str, tb_remove_name="file_checks")
    type_check(value=enforce, required_type=bool, tb_remove_name="file_checks")

    # Checks if the file does not exist
    file: pathlib.Path = pathlib.Path(file_path)

    if not file.exists():
        if enforce:
            # Sets message based on description choice.
            if file_description:
                exc_args: dict = {
                    "main_message": f"The file ({file_description}) does not exist "
                    f"in the validating file path ({file_path}).",
                    "suggested_resolution": "Ensure the file path is the correct path to your file.",
                }
            else:
                exc_args: dict = {
                    "main_message": f"The file does not exist in the validating file path ({file_path}).",
                    "suggested_resolution": "Ensure the file path is the correct path to your file.",
                }
            raise FFileNotFoundError(message_args=exc_args, tb_limit=None, tb_remove_name="file_check")
        else:
            return False
    else:
        return True
